import rfoo,thread,uuid,time,os

class BackupServer(rfoo.SSLHandler):
	def __init__(self, *args, **kwargs):
		super(BackupServer,self).__init__(*args,**kwargs)
		self.current_subdir = ''
		self._common_name = None
		self._uid = None
		self._account = None
		self._read_only = True
		self._master = False
		
		self._backend = self._context['backend']
		self._cache = self._context['cache']
		
		peer_cert = self._conn.getpeercert()
		if peer_cert != None:
			for cdn in peer_cert['subject']:
				for key,val in cdn:
					if key == 'commonName':
						self._common_name = val
					elif key == 'userId':
						self._uid = val
			if self._common_name != None:
				if self._common_name == 'public':
					self._account = 'public'
					self._read_only = False
					
				elif self._common_name == 'master':
					self._master = True
					self._account = 'public'
					self._read_only = False
					
				else:
					if self._uid != None:
						self._account = str(self._uid)
						self._read_only = False
					else:
						raise Exception('Unknown UID')
						
		else:
			self._account = 'public'
			self._read_only = True
	
	def ping(self):
		return True
	
	def getAccount(self):
		return self._account
		
	def setAccount(self, account):
		if(self._master):
			self._account = str(account)
			return True
			
		return False
		
	def listSubjects(self):
		return self._backend.get_subjects(self._account)
		
	def listBackups(self, subject, with_running = False):
		complete, running, restoring = self._backend.get_backups(self._account,subject)
		if with_running:
			return (complete, running)
		else:
			return complete
	
	def _dirFile(self, fname):
		longname = str(fname).rjust(8, '0')
		return [longname[:-4], longname]
	
	def startBackup(self, subject, btype = None, ts = None, original_subject = None, metadata = {}, job_id = None):
		if self._read_only: raise Exception('Access denied!')
		full_path, path, index, origin = self._backend.start_backup(self._account, subject, btype, ts, original_subject, metadata, job_id)
		self.current_subdir = ''
		self.backup_path = full_path
		self.origin = origin
		return (path, index, origin)
		
	def resumeBackup(self, subject, backup_identifier, origin):
		if self._read_only: raise Exception('Access denied!')
		full_path, path, index, origin = self._backend.resume_backup(self._account, subject, backup_identifier, origin)
		self.current_subdir = ''
		self.backup_path = full_path
		self.origin = origin
		return (path, index, origin)
		
	def update_backup_progress(self, backup_identifier, progress):
		if self._read_only: raise Exception('Access denied!')
		return self._backend.update_backup_progress(self._account, backup_identifier, progress)
		
	def finishBackup(self, subject, path, index, size = None):
		if self._read_only: raise Exception('Access denied!')
		return self._backend.finish_backup(self._account,path, index, size)
	
	def startRestore(self, subject, path, job_id = None):
		full_path, path, index = self._backend.start_restore(self._account, subject, path, job_id)
		self.backup_path = full_path
		return index
		
	def update_restore_progress(self,backup_identifier, job_id, progress):
		return self._backend.update_restore_progress(self._account, backup_identifier, job_id, progress)
	
	def finishRestore(self, subject, path, job_id):
		return self._backend.finish_restore(self._account, subject, path, job_id)
	
	def writeBlock(self, name, path, fname, data):
		if self._read_only: raise Exception('Access denied!')
		dirfile = self._dirFile(fname)
		if self.current_subdir != dirfile[0]:
			try:
				if not os.path.exists(os.sep.join([self.backup_path,dirfile[0]])):
					os.mkdir(os.sep.join([self.backup_path,dirfile[0]]))
				self.current_subdir = dirfile[0]
			except:
				raise
				return False
		try:
			fh = open(os.sep.join([self.backup_path]+dirfile), 'w')
			fh.write(data)
			fh.close()
		except:
			raise
			return False
		return True

	def linkBlock(self, name, path, fname):
		if self._read_only: raise Exception('Access denied!')
		if self.origin == None:
			return False
		dirfile = self._dirFile(fname)
		if self.current_subdir != dirfile[0]:
			try:
				if not os.path.exists(os.sep.join([self.backup_path,dirfile[0]])):
					os.mkdir(os.sep.join([self.backup_path,dirfile[0]]))
				self.current_subdir = dirfile[0]
			except:
				return False
		try:
			os.link(os.sep.join([self.origin]+dirfile), os.sep.join([self.backup_path]+dirfile))
			return True
		except:
			return False
	
	def readBlock(self, name, path, fname):
#		try:
		myfile = os.sep.join([self.backup_path]+self._dirFile(fname))
		fh = open(myfile, 'r')
		data = fh.read()
		fh.close()
		return data
#		except:
#			return False
	
	def change_backup_type(self, subject, backup_identifier, btype = None):
		return self._backend.change_backup_type(self._account, subject, backup_identifier, btype)
		
	def change_metadata(self, backup_identifier, data):
		return self._backend.change_metadata(self._account, backup_identifier, data)
	
	def removeBackup(self, name, path):
		if self._read_only: raise Exception('Access denied!')
		return self._backend.remove_backup(self._account,name,path)
