import cache, threading, os, time, datetime, hashlib, stat, random, re, pickle, posix

class BackupServerBackend(object):
	"""docstring for BackupServerBackend"""
	
	# TODO - delete abandoned backups
	
	def __init__(self, storage_path, cache):
		super(BackupServerBackend, self).__init__()
		self._cache = cache
		self._storage_path = storage_path
		self._data_lock = threading.Lock()
		self._to_delete = []
		self._deleting = []
		
	def maintenance_loop(self):
		max_rmdir = 3
		while True:
			try:
				with self._data_lock:
					if len(self._to_delete) > 0:
						rmdir_count = 0
						for th in threading.enumerate():
							if th.name == 'rmdir':
								rmdir_count += 1
						if rmdir_count < max_rmdir:
							# TODO - kontrolovat, ze se ze zalohy zrovna neobnovuje
							mypath = self._to_delete.pop(0)
							self._deleting.append(mypath)
							rmdir_thread = threading.Thread(target=self.__rmdir,name='rmdir',args=(mypath,))
							rmdir_thread.daemon = True
							rmdir_thread.start()
			except:
				pass
			time.sleep(1)
	
	def __rmdir(self,path):
		# TODO - sanity check na parametr, abychom nedelali rm -rf / 
		os.system('rm -rf {}'.format(path))
		with self._data_lock:
			self._deleting.remove(path)
	
	def __create_structure(self, account, subject):
		try:
			if not os.path.exists(os.path.join(self._storage_path,account,subject)):
				if not os.path.exists(os.path.join(self._storage_path,account)):
					os.mkdir(os.path.join(self._storage_path,account))
				os.mkdir(os.path.join(self._storage_path,account,subject))
			return os.path.join(self._storage_path,account,subject)
		except:
			return None
	
	def __write_index(self, mypath, index):
		try:
			index_file = open(os.path.join(mypath,'index'),'w')
			for item in index:
				index_file.write('%s,%s,%s\n' % (index[item][0], index[item][1], index[item][2]))
			index_file.close()
			return True
		except:
			raise
			return False
	
	def __get_size(self, mypath):
		size = None
		if os.path.exists(os.path.join(mypath,'.size')):
			try:
				fh = open(os.path.join(mypath,'.size'))
				size = int(fh.read())
				fh.close()
			except:
				pass
		if size == None:
			index = self.__get_index_data(os.path.join(mypath,'index'))
			if len(index) > 0:
				size = len(index)*4*1024*1024
				fh = open(os.path.join(mypath,'.size'), 'w')
				fh.write(str(size))
				fh.close()
		return size
	
	def __get_index_data(self,index_path):
		index = {}
		try:
			index_file = open(index_path, 'r')
			counter = 0
			for line in index_file:
				item = line.strip().split(',')
				item[0] = int(item[0])
				if len(item) == 1:
					if item[0] != -1:
						item.append(time.time())
					else:
						item.append(0.0)
				else:
					item[1] = float(item[1])
				if len(item) == 2:
					if item[0] != -1:
						item.append('lzo')
					else:
						item.append('')
				index[counter] = item[:3]
				counter += 1
			index_file.close()
		except:
			pass
		return index

	
	def stat(self, path, cache_timeout=30):
		cache_key = 'statcache:{}'.format(path)
		mode = self._cache.get(cache_key)
		if mode != None:
			return posix.stat_result(mode)
		else:
			try:
				mode = os.stat(path)
				self._cache.set(cache_key,list(mode),cache_timeout)
				
			except:
				mode = None
				
		return mode


	def isfile(self, path, cache_timeout=30):
		st = self.stat(path,cache_timeout)
		if st == None:
			return False
		return stat.S_ISREG(st.st_mode)

	def isdir(self, path, cache_timeout=30):
		st = self.stat(path,cache_timeout)
		if st == None:
			return False
		return stat.S_ISDIR(st.st_mode)
	
	def get_full_path(self,account,backup):
		mymatch = re.match('^(.*)\-([a-z]+)\-([0-9\-]{19})(\-[a-f0-9]+)?$', backup)
		return os.path.join(self._storage_path,account,mymatch.group(1),backup)
	
	def is_running(self,account,backup):
		"""docstring for is_running"""
		retVal = self._cache.get('running_backup:{}:{}'.format(account,backup))
		if retVal != None:
			return retVal
			
		else:
			if self.isfile(os.path.join(self.get_full_path(account,backup),'.running')):
				metadata = self.get_metadata(account, backup)
				if('__job_id__' in metadata):
					job_id = metadata['__job_id__']
					self._cache.set('running_backup:{}:{}'.format(account,backup),{'job_id': job_id, 'progress': [None, None]})
					retVal = {'job_id': job_id, 'progress': [None, None]}
					
				else:
					return False
					
			else:
				return False
			
		return retVal
		
	def is_restoring(self,account,backup, job_id = None):
		restores = self._cache.get('restoring_backup:{}:{}'.format(account,backup))
		if restores != None:
			if(job_id == None or job_id in restores):
				return True
			
		return False
		
	def get_subjects(self, account, force = False):
		cached = self._cache.get('subject_list:{}'.format(account))
		if(force):
			cached = None
		
		if(cached != None):
			return cached
			
		subjects = []
		if(self.isdir(os.path.join(self._storage_path,account))):
			ls = os.listdir(os.path.join(self._storage_path,account))
			for subject in ls:
				backup_list = self.get_backups(account, subject)
				if(len(backup_list[0]) != 0 or len(backup_list[1]) != 0):
					subjects.append(subject)
				
		self._cache.set('subject_list:{}'.format(account), subjects)
		return subjects
	
	def get_backups(self,account,subject, force = False):
		"""docstring for get_backups"""
		if(subject == None):
			complete = []
			running = []
			restoring = []
			subjects = self.get_subjects(account, force)
			for subject in subjects:
				tmp = self.get_backups(account, subject, force)
				complete += tmp[0]
				running += tmp[1]
				restoring += tmp[2]
				
			return (complete, running, restoring)
		
		cached = self._cache.get('backup_list:{}:{}'.format(account,subject))
		if(force):
			cached = None
		
		if cached != None and len(cached) == 3:
			from_cache = True
			complete, running, restoring = cached
			for backup in complete:
				if(not self.isfile(os.path.join(backup['full_path'],'index'))):
					from_cache = False
					break
					
			for backup in running:
				if(not self.is_running(account,backup['path'])):
					from_cache = False
					break
					
			for backup in restoring:
				if(not self.is_restoring(account,backup['path'], backup['job_id'])):
					from_cache = False
					break
					
		else:
			from_cache = False
			
		if from_cache:
			return cached
		
		complete = []
		running = []
		restoring = []
		
		if self.isdir(os.path.join(self._storage_path,account,subject)):
			backup_list = os.listdir(os.path.join(self._storage_path,account,subject))
			mybacks = {}
			timestamps = []
			for backup in backup_list:
				if not self.isfile(os.path.join(self._storage_path,account,subject,backup,'index')) and not self.is_running(account,backup):
					mypath = os.path.join(self._storage_path,account,subject,backup)
					with self._data_lock:
						if not mypath in (self._to_delete + self._deleting):
							self._to_delete.append(mypath)
					continue
					
				mymatch = re.match('^'+re.escape(subject)+'\-([a-z]+)\-([0-9\-]{19})(\-[a-f0-9]+)?$', backup)
				mytime = datetime.datetime(*(time.strptime(mymatch.group(2), '%Y-%m-%d-%H-%M-%S')[0:6]))
				myts = int(time.mktime(mytime.timetuple()))
				if myts not in mybacks:
					mybacks[myts] = []
				
				try:
					metadata_fh = open(os.path.join(self._storage_path,account,subject,backup,'.metadata'), 'r')
					metadata = pickle.load(metadata_fh)
					metadata_fh.close()
					
				except:
					metadata = {}	
					
				myback = {'type':mymatch.group(1), 'time':myts, 'path': backup, 'full_path': os.path.join(self._storage_path,account,subject,backup), 'base': os.path.join(self._storage_path,account,subject), 'metadata': metadata}
				mybacks[myts].append(myback)
				if self.isfile(os.path.join(self._storage_path,account,subject,backup,'index')):
					myback['size'] = self.__get_size(os.path.join(self._storage_path,account,subject,backup))
					
				timestamps.append(myts)
			timestamps.sort()
			for ts in timestamps:
				for myback in mybacks[ts]:
					running_data = self.is_running(account,myback['path'])
					if(running_data != False):
						myback.update(running_data)
						running.append(myback)
						
					else:
						complete.append(myback)
						if(self.is_restoring(account, myback['path'])):
							restores = self._cache.get('restoring_backup:{}:{}'.format(account,backup))
							if(restores == None):
								continue
							
							for job_id in restores:
								tmp = myback.copy()
								tmp['job_id'] = job_id
								tmp['progress'] = restores[job_id]
								restoring.append(tmp)
						
		self._cache.set('backup_list:{}:{}'.format(account,subject),(complete, running, restoring))
		return (complete, running, restoring)
		
	def change_metadata(self, account, backup_identifier, data):
		if(self.is_running(account,backup_identifier)):
			return False
			
		if(self.get_full_path(account,backup_identifier) in (self._to_delete + self._deleting)):
			return False
			
		try:
			metadata_fh = open(os.path.join(self.get_full_path(account,backup_identifier),'.metadata'), 'r')
			metadata = pickle.load(metadata_fh)
			metadata_fh.close()
			
		except:
			metadata = {}
			
		for name, value in data.iteritems():
			if(value == None):
				if(name in metadata):
					del metadata[name]
					
			else:
				metadata[name] = value
		
		try:		
			metadata_fh = open(os.path.join(self.get_full_path(account,backup_identifier),'.metadata'), 'w')
			pickle.dump(metadata, metadata_fh)
			metadata_fh.close()
			
		except:
			return False
		
		self.get_backups(account,subject, True)	
		return True
		
	def get_metadata(self, account, backup_identifier):
		try:
			metadata_fh = open(os.path.join(self.get_full_path(account,backup_identifier),'.metadata'), 'r')
			metadata = pickle.load(metadata_fh)
			metadata_fh.close()
			
		except:
			metadata = {}
			
		return metadata
		
	def change_backup_type(self, account, subject, backup_identifier, btype = None):
		if(self.is_running(account,backup_identifier) or self.is_restoring(account,backup_identifier)):
			return False
			
		if(os.path.join(self._storage_path, account, subject, backup_identifier) in (self._to_delete + self._deleting)):
			return False
			
		backup_list = self.get_backups(account, subject)[0]
		find = False
		for backup in backup_list:
			if(backup['path'] == backup_identifier):
				find = True
				if(btype == backup['type']):
					return True
					
				elif(btype == None):
					if(backup['type'] == 'auto'):
						btype = 'man'
						
					else:
						btype = 'auto'
						
		if(not find):
			return False
		
		new_name = re.sub('man|auto', btype, backup_identifier)
		try:	
			os.rename(os.path.join(self._storage_path, account, subject, backup_identifier), os.path.join(self._storage_path, account, subject, new_name))
			
		except:
			return False
			
		self.get_backups(account,subject, True)
		return True
		
	
	def start_backup(self, account, subject, btype = None, ts = None, original_subject = None, metadata = {}, job_id = None):
		"""docstring for start_backup"""
		if btype == None:
			btype = 'auto'
			
		if ts == None:
			btime = datetime.datetime.now()
			
		else:
			btime = datetime.datetime.fromtimestamp(ts)
			
		if original_subject == None:
			original_subject = subject
		
		if(job_id == None):
			job_id = os.urandom(8).encode('hex')
			
		complete_backups, running_backups, restoring_backup = self.get_backups(account,original_subject)
		if (complete_backups == [] and running_backups == []) or subject != original_subject:
			self.__create_structure(account,subject)
			
		if len(complete_backups) > 0:
			origin = complete_backups[-1]['full_path']
			index = self.__get_index_data(os.path.join(origin,'index'))
			
		else:
			origin = None
			index = {}
		
		backup_identifier = subject+'-'+btype+'-'+btime.strftime('%Y-%m-%d-%H-%M-%S')+'-'+hashlib.sha1(btime.strftime('%Y-%m-%d-%H-%M-%S')+str(random.random())).hexdigest()[:8]
		
		if not os.path.isdir(os.path.join(self._storage_path,account,subject,backup_identifier)):
			os.mkdir(os.path.join(self._storage_path,account,subject,backup_identifier))
		
		open(os.path.join(self._storage_path,account,subject,backup_identifier,'.running'),'w').close()
		metadata_fh = open(os.path.join(self._storage_path,account,subject,backup_identifier,'.metadata'),'w')
		metadata['__job_id__'] = job_id
		pickle.dump(metadata, metadata_fh)
		metadata_fh.close()
		self._cache.set('running_backup:{}:{}'.format(account,backup_identifier),{'job_id': job_id, 'progress': [None, None]})
		self.get_backups(account,subject, True)
		subjects = self._cache.get('subject_list:{}'.format(account))
		if(subjects != None):
			if(subject not in subjects):
				self._cache.append('subject_list:{}'.format(account), subject)
				
		return (os.path.join(self._storage_path,account,subject,backup_identifier), backup_identifier, index, origin)
	
	def resume_backup(self, account, subject, backup_identifier, origin):
		if(origin == None):
			index = {}
		else:
			index = self.__get_index_data(os.path.join(origin,'index'))
	
		job_id = self._cache.get('running_backup:{}:{}'.format(account,backup_identifier))
		if(job_id == None):
			metadata = self.get_metadata(account, backup_identifier)
			if('__job_id__' in metadata):
				job_id = metadata['__job_id__']
				self._cache.set('running_backup:{}:{}'.format(account,backup_identifier),{'job_id': job_id, 'progress': [None, None]})
				
			else:
				raise Exception('Couldn\'t retrieve job_id')
				
		self.get_backups(account,subject, True)
		return (os.path.join(self._storage_path,account,subject,backup_identifier), backup_identifier, index, origin)
	
	def update_backup_progress(self, account, backup_identifier, progress):
		job_id = self._cache.get('running_backup:{}:{}'.format(account,backup_identifier))
		if(job_id != None):
			job_id['progress'] = progress
			self._cache.set('running_backup:{}:{}'.format(account,backup_identifier),job_id)
			subject = re.match('^(.*)\-([a-z]+)\-([0-9\-]{19})(\-[a-f0-9]+)?$', backup_identifier).group(1)
			backup_list = self.get_backups(account, subject)
			for index in range(len(backup_list[1])):
				if(backup_list[1][index]['path'] == backup_identifier):
					backup_list[1][index]['progress'] = progress
					backup_list[1][index]['job_id'] =  job_id['job_id']
		
			self._cache.set('backup_list:{}:{}'.format(account,subject), backup_list)		
		
			return True
	
		return False

	
	def finish_backup(self,account,path,index,size):
		"""docstring for finish_backup"""
		result = False
		if self.is_running(account,path):
			full_path = self.get_full_path(account,path)
			subject = re.match('^(.*)\-([a-z]+)\-([0-9\-]{19})(\-[a-f0-9]+)?$', path).group(1)
			if self.__write_index(full_path, index):
				if size != None:
					fh = open(os.path.join(full_path, '.size'), 'w')
					fh.write(str(size))
					fh.close()
				
				self._cache.remove('running_backup:{}:{}'.format(account,path))
				
				os.unlink(os.path.join(full_path, '.running'))
				self.get_backups(account, subject, True)
				result = True
				
		return result
	
	def start_restore(self, account, subject, backup_identifier, job_id = None):
		# TODO - identifikuj proces obnovy
		if(job_id == None):
			job_id = os.urandom(8).encode('hex')
			progress = [None, None]
			
		else:
			resume = self._cache.get('restoring_backup:{}:{}'.format(account,backup_identifier))
			if(resume == None or job_id not in resume):
				progress = [None, None]
				
			else:
				progress = resume[job_id]
		
			
		complete = self.get_backups(account,subject)[0]
		for backup in complete:
			if backup['path'] == backup_identifier:
				with self._cache:	
					restores = self._cache.get('restoring_backup:{}:{}'.format(account,backup_identifier))
					if(restores == None):
						restores = {}
				
					restores[job_id] = progress
					self._cache.set('restoring_backup:{}:{}'.format(account,backup_identifier), restores)
					
				index = self.__get_index_data(os.path.join(backup['full_path'],'index'))
				self.get_backups(account,subject,True)
				return (backup['full_path'], backup['path'], index)
	
	def update_restore_progress(self, account, backup_identifier, job_id, progress):
		with self._cache:
			restores = self._cache.get('restoring_backup:{}:{}'.format(account,backup_identifier))
			if(restores == None or job_id not in restores):
				return False
				
			restores[job_id] = progress
			self._cache.set('restoring_backup:{}:{}'.format(account,backup_identifier), restores)
			subject = re.match('^(.*)\-([a-z]+)\-([0-9\-]{19})(\-[a-f0-9]+)?$', backup_identifier).group(1)
			backup_list = self.get_backups(account, subject)
			for index in range(len(backup_list[2])):
				if(backup_list[2][index]['path'] == backup_identifier and backup_list[2][index]['job_id'] == job_id):
					backup_list[2][index]['progress'] = progress
			
			self._cache.set('backup_list:{}:{}'.format(account,subject), backup_list)		
			return True
		
		return False
		
	
	def finish_restore(self, account, subject, backup_identifier, job_id):
		# TODO - identifikuj proces obnovy
		with self._cache:
			restores = self._cache.get('restoring_backup:{}:{}'.format(account,backup_identifier))
			if(job_id in restores):
				del restores[job_id]
				self._cache.set('restoring_backup:{}:{}'.format(account,backup_identifier), restores)
				self.get_backups(account, subject, True)
				return True
				
		return False
	
	def remove_backup(self, account, name, path):
		complete, running = self.get_backups(account,name)[0:2]
		for backup in complete:
			if backup['path'] == path:
				if os.path.exists(os.path.join(backup['full_path'],'index')):
					os.unlink(os.path.join(backup['full_path'],'index'))
					self._cache.remove('statcache:{}'.format(os.path.join(backup['full_path'],'index')))
					
				with self._data_lock:
					if not backup['full_path'] in (self._to_delete + self._deleting):
						self._to_delete.append(backup['full_path'])
				
				subjects = self._cache.get('subject_list:{}'.format(account))
				if(subjects != None):		
					if(len(complete) == 1 and len(running) == 0 and name in subjects):
						subjects.remove(name)
						self._cache.set('subject_list:{}'.format(account), subjects)
						
					elif((len(complete) > 1 or len(running) != 0) and name not in subjects):
						subjects.append(name)
						self._cache.set('subject_list:{}'.format(account), subjects)
				
				self.get_backups(account, name, True)		
				return True
	