import time,threading,re,json
from rediscluster import StrictRedisCluster
from redis import StrictRedis
 
class Cache(object):
	"""docstring for Cache"""
	def __init__(self, cache_type):
		super(Cache, self).__init__()
		self.cache_type = cache_type
		self._atomic_lock = threading.Lock()
		
	def __enter__(self):
		self._atomic_lock.acquire()
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		self._atomic_lock.release()
	
class MemoryCache(Cache):
	"""docstring for MemoryCache"""
	def __init__(self):
		super(MemoryCache, self).__init__('memory')
		self._storage = {}
		self._storage_lock = threading.Lock()
	
	def _get(self, key):
		if self._storage.has_key(key):
			itm = self._storage[key]
			if itm['timeout'] == None or itm['timeout'] > time.time():
				itm['last_access'] = time.time()
				return itm
			else:
				del(self._storage[key])
				return None
		else:
			return None
	
	def append(self, key, value):
		with self._storage_lock:
			itm = self._get(key)
			if itm == None:
				raise ValueError('Item not set!')
			if type(itm['value']) != type([]):
				raise ValueError('Item is not list!')
			itm['value'].append(value)
			self._storage[key] = itm
			
	def get(self, key):
		"""docstring for get"""
		with self._storage_lock:
			itm = self._get(key)
			if itm != None:
				return itm['value']
			else:
				return None
				
	# def get_re(self, pattern):
	# 	c_pattern = re.compile(pattern)
	# 	with self._storage_lock:
	# 		return dict(map(lambda x: (x, self._get(x)['value']), filter(lambda x: c_pattern.match(x), self._storage.keys())))
			
	
	def set(self, key, value, timeout=None):
		"""docstring for set"""
		if timeout != None and timeout <= 2592000:
			# if timeout > 30 days we assume it is unix timestamp
			timeout += time.time()
		itm = {
			'value' : value,
			'timeout': timeout,
			'last_access': None
		}
		with self._storage_lock:
			self._storage[key] = itm

		return None
	
	def remove(self,key):
		"""docstring for remove"""
		with self._storage_lock:
			if self._storage.has_key(key):
				del(self._storage[key])
		return True
		
class Redis_Cache(Cache):
	def __init__(self, nodes, prefix = 'c2p:backup:'):
		super(Redis_Cache, self).__init__('redis')
		if(len(nodes) > 1):
			self._storage = StrictRedisCluster(startup_nodes = nodes, decode_responses = True)
			
		else:
			self._storage = StrictRedis(nodes[0]['host'], nodes[0]['port'])
			
		self._prefix = prefix
		self._user_lock = threading.Lock()
		
	def get(self, key):
		key = self._prefix + key
		value = self._storage.get(key)
		if(value == None):
			return None
			
		return json.loads(self._storage.get(key))
	
	def set(self, key, value, timeout = None):
		key = self._prefix + key
		self._storage.set(key, json.dumps(value), timeout)
		
		return None
		
	def append(self, key, value):
		key = self._prefix + key
		item = self._storage.get(key)
		if(item == None):
			raise ValueError('Item not set!')
		
		item = json.loads(item)
		if(type(item) != list):
			raise ValueError('Item is not list!')
			
		item.append(value)
		ttl = self._storage.ttl(key)
		if(ttl == -1):
			ttl = None
			
		self._storage.set(key, json.dumps(item), )
		
		return None
		
	def remove(self, key, use_prefix = True):
		if(use_prefix):
			return bool(self._storage.delete(self._prefix + key))
			
		else:	
			return bool(self._storage.delete(key))
		
	def keys(self, prefix):
		return self._storage.keys(self._prefix + prefix + '*')
					