#!/usr/bin/python
from cache import MemoryCache, Redis_Cache
from backend import BackupServerBackend
from api_handler import BackupServer
from redis import Redis
import rfoo
import os,sys,ssl,threading,re

import ConfigParser


# # DONE - tohle vsechno konfigurovatelne
# SSL_CA_FILE='/etc/c2pbackup/cert/ca.crt'
# SSL_CERT_FILE='/etc/c2pbackup/cert/server.crt'
# SSL_KEY_FILE='/etc/c2pbackup/cert/server.key'
# SSL_CRL_FILE='/etc/c2pbackup/cert/crl.pem'
#
# LISTEN_HOST='0.0.0.0'
# LISTEN_PORT=50000
#
# BACKUP_STORAGE_PATH = '/mnt/vps-backup'
# # END DONE

def load_config():
	parser = ConfigParser.ConfigParser(allow_no_value=True)
	parser.read(os.path.join(os.path.dirname(__file__),'conf/defaults.ini'))
	parser.read('/etc/c2pbackup/c2pbackup.ini')
	
	return parser

def start_server():
	config = load_config()
	
	tmp = config.get('redis', 'nodes')
	tmp = map(lambda x: x.split(':'), tmp.split(';'))
	redis_nodes = []
	for node in tmp:
		if(len(node) == 1):
			if(node[0] == ''):
				continue
				
			else:
				node = {
					'host': node[0],
					'port': 6379
				}
		
		else:
			node = {
				'host': node[0],
				'port': int(node[1])
			}
			
		redis_nodes.append(node)
	
	if(len(redis_nodes) == 0):
		cache = MemoryCache()
		
	else:
		cache = Redis_Cache(redis_nodes)
	
	backend = BackupServerBackend(config.get('storage', 'path'), cache)
	if('--cache-renew' in sys.argv):
		sys.stdout.write('Renew cache data ')
		sys.stdout.flush()
		if(len(redis_nodes) == 0):
			sys.stderr.write('For redis backend only...\n')
			sys.exit(0)
		
		accounts = os.listdir(config.get('storage', 'path'))
		import time
		for account in accounts:
			if(re.match('^(([0-9]+)|(public))$',account) == None):
				continue
				
			backend.get_backups(account, None, True)
			sys.stdout.write('.')
			sys.stdout.flush()
			backup_list_keys = cache.keys('backup_list:' + account + '*')
			running_backup_keys = cache.keys('running_backup:' + account + '*')
			for key in backup_list_keys:
				subject = key.rsplit(':', 1)[1]
				if(not os.path.exists(os.path.join(config.get('storage', 'path'), account, subject))):
					cache.remove(key, False)
			
			sys.stdout.write('.')
			sys.stdout.flush()		
			for key in running_backup_keys:
				backup_identifier = key.rsplit(':', 1)[1]
				if(not os.path.isfile(os.path.join(backend.get_full_path(account,backup_identifier),'.running'))):
					cache.remove(key, False)
					
			sys.stdout.write('.')
			sys.stdout.flush()
							
		sys.stdout.write(' [  OK  ]\n')
		sys.exit(0)
	
	ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH, cafile = config.get('server', 'ssl_ca_file'))
	ssl_context.load_cert_chain(certfile = config.get('server', 'ssl_cert_file'), keyfile = config.get('server', 'ssl_key_file'))
	ssl_context.load_verify_locations(config.get('server', 'ssl_crl_file'))
	ssl_context.verify_mode = ssl.CERT_OPTIONAL
	ssl_context.verify_flags = ssl.VERIFY_DEFAULT | ssl.VERIFY_CRL_CHECK_LEAF
	
	
	th = threading.Thread(target=backend.maintenance_loop,name='maintenance_loop')
	th.daemon = True
	th.start()
	
	context = {
		'cache': cache,
		'backend': backend,
	}
	
	rfoo.SSLServer(BackupServer,context,ssl_context=ssl_context, timeout = config.getint('server', 'timeout')).start(host = config.get('server', 'listen'), port = config.getint('server', 'port'))
